// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Package iosdeps is a just a list of the packages we import on iOS, to let us
// test that our transitive closure of dependencies on iOS doesn't accidentally
// grow too large, as we've historically been memory constrained there.
//
// It is intended to mirror the imports of the ipn-go-bridge package in the
// private "corp" repository (the Go side of the iOS / macOS app).
package iosdeps

import (
	_ "bufio"
	_ "bytes"
	_ "crypto"
	_ "crypto/ecdsa"
	_ "crypto/elliptic"
	_ "crypto/sha256"
	_ "encoding/base64"
	_ "encoding/json"
	_ "errors"
	_ "fmt"
	_ "io"
	_ "log"
	_ "math"
	_ "net"
	_ "net/http"
	_ "net/netip"
	_ "net/url"
	_ "os"
	_ "os/signal"
	_ "path/filepath"
	_ "runtime"
	_ "runtime/debug"
	_ "slices"
	_ "strconv"
	_ "strings"
	_ "sync"
	_ "sync/atomic"
	_ "syscall"
	_ "time"
	_ "unsafe"

	_ "github.com/tailscale/wireguard-go/device"
	_ "github.com/tailscale/wireguard-go/tun"
	_ "golang.org/x/sys/unix"
	_ "tailscale.com/client/tailscale/apitype"
	_ "tailscale.com/drive/driveimpl"
	_ "tailscale.com/envknob"
	_ "tailscale.com/feature/condregister"
	_ "tailscale.com/feature/syspolicy"
	_ "tailscale.com/feature/taildrop"
	_ "tailscale.com/hostinfo"
	_ "tailscale.com/ipn"
	_ "tailscale.com/ipn/ipnauth"
	_ "tailscale.com/ipn/ipnlocal"
	_ "tailscale.com/ipn/localapi"
	_ "tailscale.com/logpolicy"
	_ "tailscale.com/logtail"
	_ "tailscale.com/logtail/filch"
	_ "tailscale.com/net/dns"
	_ "tailscale.com/net/netmon"
	_ "tailscale.com/net/netutil"
	_ "tailscale.com/net/tsaddr"
	_ "tailscale.com/net/tsdial"
	_ "tailscale.com/net/tshttpproxy"
	_ "tailscale.com/net/tstun"
	_ "tailscale.com/paths"
	_ "tailscale.com/safesocket"
	_ "tailscale.com/tsd"
	_ "tailscale.com/types/empty"
	_ "tailscale.com/types/key"
	_ "tailscale.com/types/lazy"
	_ "tailscale.com/types/logger"
	_ "tailscale.com/types/logid"
	_ "tailscale.com/types/netmap"
	_ "tailscale.com/util/clientmetric"
	_ "tailscale.com/util/dnsname"
	_ "tailscale.com/util/eventbus"
	_ "tailscale.com/util/must"
	_ "tailscale.com/util/set"
	_ "tailscale.com/util/syspolicy"
	_ "tailscale.com/util/syspolicy/pkey"
	_ "tailscale.com/util/syspolicy/setting"
	_ "tailscale.com/util/syspolicy/source"
	_ "tailscale.com/version"
	_ "tailscale.com/wgengine"
	_ "tailscale.com/wgengine/netstack"
	_ "tailscale.com/wgengine/router"
)
