// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Code generated by gen.go; DO NOT EDIT.

//go:build ts_omit_unixsocketidentity

package buildfeatures

// HasUnixSocketIdentity is whether the binary was built with support for modular feature "differentiate between users accessing the LocalAPI over unix sockets (if omitted, all users have full access)".
// Specifically, it's whether the binary was NOT built with the "ts_omit_unixsocketidentity" build tag.
// It's a const so it can be used for dead code elimination.
const HasUnixSocketIdentity = false
