// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

//go:build gokrazy

package dns

const (
	resolvConf = "/tmp/resolv.conf"
	backupConf = "/tmp/resolv.pre-tailscale-backup.conf"
)
