// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Package posture contains functions to query the local system
// state for managed posture checks.
package posture
