// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Package linkspeed registers support for setting the TUN link speed on Linux,
// to better integrate with system monitoring tools.
package linkspeed
