// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Code generated by gen.go; DO NOT EDIT.

//go:build ts_omit_capture

package buildfeatures

// HasCapture is whether the binary was built with support for modular feature "Packet capture".
// Specifically, it's whether the binary was NOT built with the "ts_omit_capture" build tag.
// It's a const so it can be used for dead code elimination.
const HasCapture = false
