// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Code generated by gen.go; DO NOT EDIT.

//go:build !ts_omit_portmapper

package buildfeatures

// HasPortMapper is whether the binary was built with support for modular feature "NAT-PMP/PCP/UPnP port mapping support".
// Specifically, it's whether the binary was NOT built with the "ts_omit_portmapper" build tag.
// It's a const so it can be used for dead code elimination.
const HasPortMapper = true
