// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

//go:build !plan9

// Package apis contains a constant to name the Tailscale Kubernetes Operator's schema group.
package apis

const GroupName = "tailscale.com"
