// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

//go:build (ts_aws || (linux && (arm64 || amd64) && !android)) && !ts_omit_aws

package awsparamstore

import _ "tailscale.com/feature/awsparamstore"
