// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Package identityfederation registers support for authkey resolution
// via identity federation if it's not disabled by the
// ts_omit_identityfederation build tag.
package identityfederation
