// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Package geo provides functionality to represent and process geographical
// locations on a spherical Earth.
package geo
