// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Package racebuild exports a constant about whether the current binary
// was built with the race detector.
package racebuild
