// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

//go:build !ts_omit_identityfederation

package identityfederation

import _ "tailscale.com/feature/identityfederation"
