// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Package vms does VM-based integration/functional tests by using
// qemu and a bank of pre-made VM images.
package vms
