// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Code generated by gen.go; DO NOT EDIT.

//go:build ts_omit_ipnbus

package buildfeatures

// HasIPNBus is whether the binary was built with support for modular feature "IPN notification bus (watch-ipn-bus) support, used by GUIs, debugging, and nicer 'tailscale up' support".
// Specifically, it's whether the binary was NOT built with the "ts_omit_ipnbus" build tag.
// It's a const so it can be used for dead code elimination.
const HasIPNBus = false
