// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Code generated by gen.go; DO NOT EDIT.

//go:build !ts_omit_useproxy

package buildfeatures

// HasUseProxy is whether the binary was built with support for modular feature "Support using system proxies as specified by env vars or the system configuration to reach Tailscale servers.".
// Specifically, it's whether the binary was NOT built with the "ts_omit_useproxy" build tag.
// It's a const so it can be used for dead code elimination.
const HasUseProxy = true
