// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

//go:build ts_omit_aws

package omit

// AWS is whether AWS support should be omitted from the build.
const AWS = true
