// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

//go:build ignore

// The tsshd binary was an experimental SSH server that accepts connections
// from anybody on the same Tailscale network.
//
// Its functionality moved into tailscaled.
//
// See https://github.com/tailscale/tailscale/issues/3802
package main
