// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Package useproxy registers support for using proxies
// if it's not disabled via the ts_omit_useproxy build tag.
package useproxy
