// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

package tsconst

// WebListenPort is the static port used for the web client when run inside
// tailscaled. (5252 are the numbers above the letters "TSTS" on a qwerty
// keyboard.)
const WebListenPort = 5252
