// Copyright (c) Tailscale Inc & contributors
// SPDX-License-Identifier: BSD-3-Clause

// Package awsparamstore conditionally registers the awsparamstore feature for
// resolving secrets from AWS Parameter Store.
package awsparamstore
